
import patsy
import stanity
import pandas as pd
import numpy as np


def fit_stan_survival_model(df, formula, event_col, model_code = None, file=None,
                                           model_cohort = 'survival model', 
                             time_col = None,
                             sample_id_col = None, sample_col = None,
                             group_id_col = None, group_col = None,
                             timepoint_id_col = None, timepoint_end_col = None,
                             make_inits = None, stan_data = None,
                             grp_coef_type = None, FIT_FUN = stanity.fit,
                             drop_intercept = True,
                             *args, **kwargs):
    """This function prepares inputs appropriate for stan model model code, and fits that model using Stan.

    Args:
       df (pandas DataFrame):  The data frame containing input data to Survival model.
       formula (chr): Patsy formula to use for covariates. E.g 'met_status + pd_l1'
       event_col (chr): name of column containing event status. Will be coerced to int
       model_code (chr): stan model code to use.
       file (chr): path to stan file (if model_code not given).

    Kwargs:
       model_cohort (chr): description of this model fit, to be used when plotting or summarizing output
       time_col (chr): name of column containing event time -- used for parameteric (Weibull) model
       sample_id_col (chr): name of column containing numeric sample ids (1-indexed & sequential)
       sample_col (chr): name of column containing sample descriptions - will be converted to an ID
       group_id_col (chr): name of column containing numeric group ids (1-indexed & sequential)
       group_col (chr): name of column containing group descriptions - will be converted to an ID 
       timepoint_id_col (chr): name of column containing timepoint ids (1-indexed & sequential)
       timepoint_end_col (chr): name of column containing end times for each timepoint (will be converted to an ID)
       stan_data (dict): extra params passed to stan data object
       grp_coef_type (chr): type of group coef specified, if using a varying-coef model
              Can be one of:
              - 'None' (default): guess group coef orientation from data. 
                                  Works except in case where M (num covariates) == G (num groups)
              - 'matrix': grp_beta defined as `matrix[M, G] grp_beta;`
              - 'vector-of-vectors': grp_beta defined as `vector[M] grp_beta[G];`
       drop_intercept (bool): whether to drop the intercept term from the model matrix (default: True)

    Returns:
       dictionary of results objects.  Contents::

          df: Pandas data frame containing input data, filtered to non-missing obs & with ID variables created
          x_df: Covariate matrix passed to Stan
          x_names: Column names for the covariate matrix passed to Stan
          data: List passed to Stan - contains dimensions, etc.
          fit: pystan fit object returned from Stan call
          coefs: posterior draws for coefficient values
          loo: psis-loo object returned for fit model. Used for model comparison & summary
          model_cohort: description of this model and/or cohort on which the model was fit

    Raises:
       AttributeError, KeyError

    Generic helper function for fitting variety of survival models using Stan.

    Example:

    >>> testfit = fit_stan_survival_model(
                model_file = stanmodels.stan.pem_survival_model,
                formula = '~ met_status + pd_l1',
                df = dflong,
                sample_col = 'patient_id',
                timepoint_end_col = 'end_time',
                event_col = 'end_failure',
                model_cohort = 'PEM survival model',
                iter = 30000,
                chains = 4,
            )
    >>> print(testfit['fit'])
    >>> seaborn.boxplot(x = 'value', y = 'variable', data = testfit['coefs'])

    """

    if model_code is None:
        if file is None:
            raise AttributeError('Either model_code or file is required.')
    
    ## input covariates given formula
    x_df = patsy.dmatrix(formula,
                          df,
                          return_type='dataframe'
                          )
    
    ## construct data frame with all necessary columns
    ## limit to non-missing data 
    ## (if necessary) transform columns to ids
    other_cols = [event_col, time_col,
                  group_id_col, group_col,
                  timepoint_id_col, timepoint_end_col,
                  sample_id_col, sample_col] ## list of possible columns to keep
    
    other_cols = list(set(other_cols))
    other_cols.remove(None)
    
    if other_cols and len(other_cols)>0:
        ## filter other inputs to non-missing observations on input covariates
        df_nonmiss = x_df.join(df[other_cols]).dropna()
    else:
        df_nonmiss = x_df

    if len(x_df.columns)>1 and drop_intercept:
        x_df = x_df.ix[:, x_df.columns != 'Intercept']

    ## prep input dictionary to pass to stan.fit
    survival_model_input_data = {
        'N': len(df_nonmiss.index),
        'x': x_df.as_matrix(),
        'event': df_nonmiss[event_col].values.astype(int),
        'M': len(x_df.columns),
    }

    if time_col:
        survival_model_input_data['y'] = df_nonmiss[time_col].values

    ## construct timepoint ID vars & add to input data
    if timepoint_end_col and not(timepoint_id_col):
        timepoint_id_col = 'timepoint_id'
        df_nonmiss[timepoint_id_col] = df_nonmiss[timepoint_end_col].astype('category').cat.codes + 1

    if timepoint_id_col:
        unique_timepoints = _prep_timepoint_dataframe(df_nonmiss,
                                                      timepoint_id_col = timepoint_id_col,
                                                      timepoint_end_col = timepoint_end_col
                                                      )
        timepoint_input_data = {
            't_dur': unique_timepoints['t_dur'],
            't_obs': unique_timepoints[timepoint_end_col],
            't': df_nonmiss[timepoint_id_col].values.astype(int),
            'T': len(df_nonmiss[timepoint_id_col].unique())
        }
        survival_model_input_data = dict(survival_model_input_data, **timepoint_input_data)

    if timepoint_end_col:
        # not required for all models, leave in for legacy
        survival_model_input_data['obs_t'] = df_nonmiss[timepoint_end_col].values.astype(int)

    ## construct sample ID var & add to input data
    if sample_col and not(sample_id_col):
        sample_id_col = 'sample_id'
        df_nonmiss[sample_id_col] = df_nonmiss[sample_col].astype('category').cat.codes + 1

    if sample_id_col:
        sample_input_data = {
            's': df_nonmiss[sample_id_col].values.astype(int),
            'S': len(df_nonmiss[sample_id_col].unique())
        }
        survival_model_input_data = dict(survival_model_input_data, **sample_input_data)
 
    ## construct group ID var & add to input data
    if group_col and not(group_id_col):
        group_id_col = 'group_id'
        df_nonmiss[group_id_col] = df_nonmiss[group_col].astype('category').cat.codes + 1

    if group_id_col:
        survival_model_input_data['g'] = df_nonmiss[group_id_col].values.astype(int)
        survival_model_input_data['G'] = len(df_nonmiss[group_id_col].unique())
        
    if stan_data:
        survival_model_input_data = dict(survival_model_input_data, **stan_data)
    
    if make_inits:
        kwargs = dict(kwargs, init = make_inits(survival_model_input_data))
    
    survival_fit = FIT_FUN(
        model_code = model_code,
        file = file,
        data = survival_model_input_data,
        *args,
        **kwargs
    )
    
    try:
        beta_coefs = pd.DataFrame(
            survival_fit.extract()['beta'],
            columns = x_df.columns
        )
        beta_coefs.reset_index(0, inplace = True)
        beta_coefs = beta_coefs.rename(columns = {'index':'iter'})
        beta_coefs = pd.melt(beta_coefs, id_vars = ['iter'])
        beta_coefs['exp(beta)'] = np.exp(beta_coefs['value'])
        beta_coefs['model_cohort'] = model_cohort
    except:
        beta_coefs = None
    
    ## prep by-group coefs if group specified
    if group_id_col:
        try:
            if group_col:
                grp_names = df_nonmiss.loc[
                    ~df_nonmiss[[group_id_col]].duplicated()].sort_values(group_id_col)[group_col].values
            else:
                grp_names = df_nonmiss.loc[
                    ~df_nonmiss[[group_id_col]].duplicated()].sort_values(group_id_col)[group_id_col].values
            grp_coefs = _extract_grp_coefs(survival_fit=survival_fit,
                                           element='grp_beta',
                                           grp_coef_type=grp_coef_type,
                                           grp_names=grp_names,
                                           columns=x_df.columns,
                                           input_data=survival_model_input_data,
                                           model_cohort=model_cohort
                                          )
        except:
            grp_coefs = None
    else:
        grp_coefs = beta_coefs
        if grp_coefs is not None:
            grp_coefs['group'] = 'Overall'

    try:
        loo = stanity.psisloo(survival_fit.extract()['log_lik'])
    except:
        loo = None
    
    if not sample_id_col:
        sample_id_col = None
    if not sample_col:
        sample_col = None
    if not timepoint_id_col:
        timepoint_id_col = None
    if not timepoint_end_col:
        timepoint_end_col = None

    return {
        'df': df_nonmiss,
        'x_df': x_df,
        'x_names': x_df.columns,
        'data': survival_model_input_data,
        'fit': survival_fit,
        'coefs': beta_coefs,
        'grp_coefs': grp_coefs,
        'loo': loo,
        'model_cohort': model_cohort,
        'df_all': df,
        'sample_col': sample_col,
        'sample_id_col': sample_id_col,
        'timepoint_id_col': timepoint_id_col,
        'timepoint_end_col': timepoint_end_col,
    }


def _extract_grp_coefs(survival_fit, element, grp_coef_type, grp_names, columns, input_data, model_cohort):
    """ Helper function to extract grp coefs summary data
    """
    grp_coefs_extract = survival_fit.extract()[element]
    
    ## try to guess shape of group-betas
    if not(grp_coef_type):
        grp_coef_type = _guess_grp_coef_type(extract=grp_coefs_extract,
                                             input_data=input_data)
        
    ## process group_coefs according to type
    if grp_coef_type == 'matrix':
        try:
            grp_coefs_data = _format_grp_coefs_matrix(extract=grp_coefs_extract,
                                                      columns=columns,
                                                      grp_names=grp_names
                                                     )
        except:
            raise Exception('unable to format grp coefs as matrix')
    elif grp_coef_type == 'vector-of-vectors':
        try:
            grp_coefs_data = _format_grp_coefs_vectors(extract=grp_coefs_extract,
                                                       columns=columns,
                                                       grp_names=grp_names
                                                      )
        except:
            raise Exception('unable to format grp coefs as vector-of-vectors')
    elif grp_coef_type == 'unknown':
        print("warning: unable to determine group-coef orientation. Try using arg `grp_coef_type`")
        return(None)
    else:
        print("Invalid `grp_coef_type` -- must be one of 'vector-of-vectors' or 'matrix'")
        print("Skipping grp coef extraction for now.")
        return(None)
    
    # process/format grp_coefs data
    grp_coefs = pd.melt(grp_coefs_data, id_vars=['group','iter'])
    grp_coefs['exp(beta)'] = np.exp(grp_coefs['value'])
    grp_coefs['group'] = grp_coefs.group.astype('category')
    grp_coefs['model_cohort'] = model_cohort
    return(grp_coefs)

def _format_grp_coefs_matrix(extract, columns, grp_names):
    """ Helper function for format grp_coefs data if in `matrix[M, G]` form
    """
    grp_coefs_data = list()
    i = 0
    for grp in grp_names:
        grp_data = pd.DataFrame(extract[:,:,i], columns = columns)
        grp_data.reset_index(inplace=True)
        grp_data.rename(columns={'index':'iter'}, inplace=True)
        grp_data['group'] = grp
        grp_coefs_data.append(grp_data)
        i = i+1
    return(pd.concat(grp_coefs_data))

def _format_grp_coefs_vectors(extract, columns, grp_names):
    """ Helper function for format grp_coefs data if in `vector[M] grp_beta[G]` form
    """
    grp_coefs_data = list()
    i = 0
    for grp in grp_names:
        grp_data = pd.DataFrame(extract[:,i,:], columns = columns)
        grp_data.reset_index(inplace=True)
        grp_data.rename(columns={'index':'iter'}, inplace=True)
        grp_data['group'] = grp
        grp_coefs_data.append(grp_data)
        i = i+1
    return(pd.concat(grp_coefs_data))
    

def _guess_grp_coef_type(extract, input_data):
    """ helper function to determine grp_coefs type from shape of returned object
    """
    if input_data['M'] == input_data['G']:
        # unable to determine shape if M == G
        grp_coef_type = 'unknown'
    elif extract.shape[1] == input_data['G']:
        grp_coef_type = 'vector-of-vectors'
    elif extract.shape[2] == input_data['G']:
        grp_coef_type = 'matrix'
    return grp_coef_type

def _prep_timepoint_dataframe(df,
                              timepoint_end_col,
                              timepoint_id_col = None
                              ):
    """ Helper function to take a set of timepoints 
        in observation-level dataframe & return 
        formatted timepoint_id, end_time, duration 

        Returns
        ---------
        pandas dataframe with one record per timepoint_id
            where timepoint_id is the index
            sorted on the index, increasing

    """
    time_df = df.copy()
    time_df.sort_values(timepoint_end_col, inplace=True)
    if not(timepoint_id_col):
        timepoint_id_col = 'timepoint_id'
        time_df[timepoint_id_col] = time_df[timepoint_end_col].astype('category').cat.codes + 1        
    time_df.dropna(how='any', subset=[timepoint_id_col, timepoint_end_col], inplace=True)
    time_df = time_df.loc[:,[timepoint_id_col, timepoint_end_col]].drop_duplicates()
    time_df[timepoint_end_col] = time_df[timepoint_end_col].astype(np.float32)
    time_df.set_index(timepoint_id_col, inplace=True, drop=True)
    time_df.sort_index(inplace=True)
    t_durs = time_df.diff(periods=1)
    t_durs.rename(columns = {timepoint_end_col: 't_dur'}, inplace=True)
    time_df = time_df.join(t_durs)
    time_df.fillna(inplace=True, value=time_df.loc[1, timepoint_end_col])
    return(time_df)


def extract_grp_baseline_hazard(results, timepoint_id_col = 'timepoint_id', timepoint_end_col = 'end_time'):
    """ If model results contain a grp_baseline object, extract & summarize it
    """

    ## TODO check if results are by-group
    ## TODO check if baseline hazard is computable
    grp_baseline_extract = results['fit'].extract()['grp_baseline']
    coef_group_names = results['grp_coefs']['group'].unique()
    i = 0
    grp_baseline_data = list()
    for grp in coef_group_names:
        grp_base = pd.DataFrame(grp_baseline_extract[:,:,i])
        grp_base_coefs = pd.melt(grp_base, var_name=timepoint_id_col, value_name='baseline_hazard')
        grp_base_coefs['group'] = grp
        grp_baseline_data.append(grp_base_coefs)
        i = i+1
    grp_baseline_coefs = pd.concat(grp_baseline_data)
    end_times = _extract_timepoint_end_times(results, timepoint_id_col = timepoint_id_col, timepoint_end_col = timepoint_end_col)
    bs_data = pd.merge(grp_baseline_coefs, end_times, on = timepoint_id_col) 
    return(bs_data)

def _extract_timepoint_end_times(results, timepoint_end_col = 'end_time', timepoint_id_col = 'timepoint_id'):
    df_nonmiss = results['df']
    end_times = df_nonmiss.loc[~df_nonmiss[[timepoint_id_col]].duplicated()].sort_values(timepoint_id_col)[[timepoint_end_col, timepoint_id_col]]
    return(end_times)

def extract_baseline_hazard(results, element='baseline', timepoint_id_col = 'timepoint_id', timepoint_end_col = 'end_time'):
    """ If model results contain a baseline object, extract & summarize it
    """
    ## TODO check if baseline hazard is computable
    baseline_extract = results['fit'].extract()[element]
    baseline_coefs = pd.DataFrame(baseline_extract)
    bs_coefs = pd.melt(baseline_coefs, var_name = timepoint_id_col, value_name = 'baseline_hazard')
    end_times = _extract_timepoint_end_times(results, timepoint_id_col = timepoint_id_col, timepoint_end_col = timepoint_end_col)
    bs_data = pd.merge(bs_coefs, end_times, on = timepoint_id_col)
    bs_data['model_cohort'] = results['model_cohort']
    return(bs_data)

## convert wide survival data to long format
def prep_data_long_surv(df, time_col, event_col):
    ''' convert wide survival data to long format
    '''
    ## identify distinct failure/censor times
    failure_times = df[time_col].unique()
    ftimes = pd.DataFrame({'end_time': failure_times, 'key':1})
    
    ## cross join failure times with each observation
    df['key'] = 1
    dflong = pd.merge(df, ftimes, on = 'key')
    
    ## identify end-time & end-status for each sample*failure time
    def gen_end_failure(row):
        if row[time_col] > row['end_time']:
            ## event not yet occurred (time_col is after this timepoint)
            return False
        if row[time_col] == row['end_time']:
            ## event during (==) this timepoint
            return row[event_col]
        if row[time_col] < row['end_time']:
            ## event already occurred (time_col is before this timepoint)
            return np.nan

    dflong['end_failure'] = dflong.apply(lambda row: gen_end_failure(row), axis = 1)
    
    ## confirm total number of non-censor events hasn't changed
    if not(sum(dflong.end_failure.dropna()) == sum(df[event_col].dropna())):
        print('Warning: total number of events has changed from {0} to {1}'.format(sum(df[event_col]), sum(dflong.end_failure)))

    
    ## remove timepoints after failure/censor event
    dflong = dflong.query('end_time <= {0}'.format(time_col))
    return(dflong)

def make_weibull_survival_model_inits(stan_input_dict):
    def f():
        m = {
            'tau_s_raw': abs(np.random.normal(0, 1)),
            'tau_raw': abs(np.random.normal(0, 1, stan_input_dict['M'])),
            'alpha_raw': np.random.normal(0, 0.1),
            'beta_raw': np.random.normal(0, 1, stan_input_dict['M']),
            'mu': np.random.normal(0, 1),
        }
        return m
    return f





import matplotlib as mpl
mpl.use('Agg')
import survivalstan
from stancache import stancache
import numpy as np
from functools import partial
from nose.tools import ok_
num_iter = 500
from .test_datasets import sim_test_dataset

model_code = '''
functions {
    int count_value(vector a, real val) {
        int s;
        s = 0;
        for (i in 1:num_elements(a)) 
            if (a[i] == val) 
                s = s + 1;
        return s;
    }

  // Defines the log survival
  real surv_gamma_lpdf (vector t, vector d, real shape, vector rate, int num_cens, int num_obs) {
    vector[2] log_lik;
    int idx_obs[num_obs];
    int idx_cens[num_cens];
    real prob;
    int i_cens;
    int i_obs;
    i_cens = 1;
    i_obs = 1;
    for (i in 1:num_elements(t)) {
        if (d[i] == 1) {
            idx_obs[i_obs] = i;
            i_obs = i_obs+1;
        }
        else {
            idx_cens[i_cens] = i;
            i_cens = i_cens+1;
        }
    }
    print(idx_obs);
    log_lik[1] = gamma_lpdf(t[idx_obs] | shape, rate[idx_obs]);
    log_lik[2] = gamma_lccdf(t[idx_cens] | shape, rate[idx_cens]);
    prob = sum(log_lik);
    return prob;
  }
}
data {
  int N;                            // number of observations
  vector<lower=0>[N] y;             // observed times
  vector<lower=0,upper=1>[N] event; // censoring indicator (1=observed, 0=censored)
  int M;                            // number of covariates
  matrix[N, M] x;                   // matrix of covariates (with n rows and H columns)
}
transformed data {
  int num_cens;
  int num_obs;
  num_obs = count_value(event, 1);
  num_cens = N - num_obs;
}
parameters {
  vector[M] beta;         // Coefficients in the linear predictor (including intercept)
  real<lower=0> alpha;    // shape parameter
}
transformed parameters {
  vector[N] linpred;
  vector[N] mu;
  linpred = x*beta;
  mu = exp(linpred);
}
model {
  alpha ~ gamma(0.01,0.01);
  beta ~ normal(0,5);
  y ~ surv_gamma(event, alpha, mu, num_cens, num_obs);
}
'''
make_inits = None

def test_null_model(**kwargs):
    ''' Test weibull survival model on simulated dataset
    '''
    d = sim_test_dataset()
    testfit = survivalstan.fit_stan_survival_model(
        model_cohort = 'test model',
        model_code = model_code,
        df = d,
        time_col = 't',
        event_col = 'event',
        formula = '~ 1',
        iter = num_iter,
        chains = 2,
        seed = 9001,
        make_inits = make_inits,
        FIT_FUN = stancache.cached_stan_fit,
        drop_intercept = False,
        **kwargs
        )
    ok_('fit' in testfit)
    ok_('coefs' in testfit)
    survivalstan.utils.plot_coefs([testfit])
    survivalstan.utils.plot_coefs([testfit], trans=np.exp)
    return(testfit)

def test_model(**kwargs):
    ''' Test weibull survival model on simulated dataset
    '''
    d = sim_test_dataset()
    testfit = survivalstan.fit_stan_survival_model(
        model_cohort = 'test model',
        model_code = model_code,
        df = d,
        time_col = 't',
        event_col = 'event',
        formula = '~ age + sex',
        iter = num_iter,
        chains = 2,
        seed = 9001,
        make_inits = make_inits,
        FIT_FUN = stancache.cached_stan_fit,
        drop_intercept = False,
        **kwargs
        )
    ok_('fit' in testfit)
    ok_('coefs' in testfit)
    survivalstan.utils.plot_coefs([testfit])
    survivalstan.utils.plot_coefs([testfit], trans=np.exp)
    return(testfit)





import pkg_resources, os

resource_package = __name__  ## Could be any module/package name.

_pem_survival_unstructured_path = os.path.join('stan', 'pem_survival_model_unstructured.stan')
_pem_survival_randomwalk_path = os.path.join('stan', 'pem_survival_model_randomwalk.stan')
_pem_survival_gamma_path = os.path.join('stan', 'pem_survival_model_gamma.stan')
_pem_survival_timevarying_path = os.path.join('stan', 'pem_survival_model_tvc.stan')
_weibull_survival_path = os.path.join('stan', 'weibull_survival_model.stan')
_exp_survival_path = os.path.join('stan', 'exp_survival_model.stan')

# varying-coefs models
_weibull_survival_varcoef_path = os.path.join('stan', 'weibull_survival_model_varying_coefs.stan')
_pem_survival_varcoef_path = os.path.join('stan', 'pem_survival_model_unstructured_varcoef.stan')

pem_survival_model = pkg_resources.resource_string(
	resource_package, _pem_survival_unstructured_path).decode("utf-8") 
pem_survival_model_unstructured = pkg_resources.resource_string(
	resource_package, _pem_survival_unstructured_path).decode("utf-8") 
pem_survival_model_randomwalk = pkg_resources.resource_string(
	resource_package, _pem_survival_randomwalk_path).decode("utf-8") 
pem_survival_model_gamma = pkg_resources.resource_string(
	resource_package, _pem_survival_gamma_path).decode("utf-8") 
pem_survival_model_timevarying = pkg_resources.resource_string(
	resource_package, _pem_survival_timevarying_path).decode("utf-8") 
weibull_survival_model = pkg_resources.resource_string(
	resource_package, _weibull_survival_path).decode("utf-8") 
exp_survival_model = pkg_resources.resource_string(
	resource_package, _exp_survival_path).decode("utf-8")

# varying-coefs models
pem_survival_model_varying_coefs = pkg_resources.resource_string(
        resource_package, _pem_survival_varcoef_path).decode("utf-8")
weibull_survival_model_varying_coefs = pkg_resources.resource_string(
        resource_package, _weibull_survival_varcoef_path).decode("utf-8")




import matplotlib as mpl
mpl.use('Agg')
import survivalstan
from stancache import stancache
import numpy as np
from nose.tools import ok_
from functools import partial
num_iter = 1000
from .test_datasets import sim_test_dataset

model_code = survivalstan.models.exp_survival_model
make_inits = None

def test_null_model_sim(**kwargs):
    ''' Test survival model on simulated dataset
    '''
    d = sim_test_dataset()
    testfit = survivalstan.fit_stan_survival_model(
        model_cohort = 'test model',
        model_code = model_code,
        df = d,
        time_col = 't',
        event_col = 'event',
        formula = '~ 1',
        iter = num_iter,
        chains = 2,
        FIT_FUN = stancache.cached_stan_fit,
        seed = 9001,
        make_inits = make_inits,
        **kwargs
        )
    ok_('fit' in testfit)
    ok_('coefs' in testfit)
    ok_('loo' in testfit)
    survivalstan.utils.plot_coefs([testfit])
    survivalstan.utils.plot_coefs([testfit], trans=np.exp)
    return(testfit)


def test_model_sim(**kwargs):
    ''' Test survival model on simulated dataset
    '''
    d = sim_test_dataset()
    testfit = survivalstan.fit_stan_survival_model(
        model_cohort = 'test model',
        model_code = model_code,
        df = d,
        time_col = 't',
        event_col = 'event',
        formula = '~ age + sex',
        iter = num_iter,
        chains = 2,
        FIT_FUN = stancache.cached_stan_fit,
        seed = 9001,
        make_inits = make_inits,
        **kwargs
        )
    ok_('fit' in testfit)
    ok_('coefs' in testfit)
    ok_('loo' in testfit)
    survivalstan.utils.plot_coefs([testfit])
    survivalstan.utils.plot_coefs([testfit], trans=np.exp)
    return(testfit)









import matplotlib as mpl
mpl.use('Agg')
import survivalstan
from stancache import stancache
import numpy as np
from nose.tools import ok_
from functools import partial
num_iter = 500
from .test_datasets import load_test_dataset_long, sim_test_dataset_long

model_code = survivalstan.models.pem_survival_model_timevarying
make_inits = None

def test_pem_model(**kwargs):
    ''' Test survival model on test dataset
    '''
    dlong = sim_test_dataset_long()
    testfit = survivalstan.fit_stan_survival_model(
        model_cohort = 'test model',
        model_code = model_code,
        df = dlong,
        sample_col = 'index',
        timepoint_end_col = 'end_time',
        event_col = 'end_failure',
        formula = '~ age + sex',
        iter = num_iter,
        chains = 2,
        seed = 9001,
        make_inits = make_inits,
        FIT_FUN = stancache.cached_stan_fit,
        **kwargs
        )
    ok_('fit' in testfit)
    ok_('coefs' in testfit)
    ok_('loo' in testfit)
    survivalstan.utils.plot_coefs([testfit])
    survivalstan.utils.plot_coefs([testfit], trans=np.exp)
    survivalstan.utils.plot_coefs([testfit], trans=np.exp, element='grp_coefs')
    survivalstan.utils.plot_coefs([testfit], element='baseline')
    survivalstan.utils.plot_pp_survival([testfit])
    time_betas = survivalstan.utils.extract_time_betas([testfit])
    return(testfit) 

def test_pem_null_model(force=True, **kwargs):
    ''' Test NULL survival model on flchain dataset
    '''
    dlong = sim_test_dataset_long()
    testfit = survivalstan.fit_stan_survival_model(
        model_cohort = 'test model',
        model_code = model_code,
        df = dlong,
        sample_col = 'index',
        timepoint_end_col = 'end_time',
        event_col = 'end_failure',
        formula = '~ 1',
        iter = num_iter,
        chains = 2,
        seed = 9001,
        make_inits = make_inits,
        FIT_FUN = stancache.cached_stan_fit,
        **kwargs
        )
    ok_('fit' in testfit)
    ok_('coefs' in testfit)
    ok_('loo' in testfit)
    survivalstan.utils.plot_coefs([testfit])
    survivalstan.utils.plot_coefs([testfit], trans=np.exp)
    survivalstan.utils.plot_coefs([testfit], trans=np.exp, element='grp_coefs')
    survivalstan.utils.plot_coefs([testfit], element='baseline')
    survivalstan.utils.plot_pp_survival([testfit])
    return(testfit)






"""

   Functions to simulate failure-time data 
   for testing & model checking purposes

"""

import numpy as np
import pandas as pd
import patsy

def sim_data_exp(N, censor_time, rate):
    """
      simulate true lifetimes (t) according to exponential model

      Parameters
      -----------

        N: (int) number of observations 
        censor_time: (float) uniform censor time for each observation
		rate: (float, positive) hazard rate used to parameterize failure times

	  Returns
	  -------

	  pandas DataFrame with N observations, and 3 columns:
	    - true_t: "actual" simulated failure time
	    - t: observed failure/censor time, given censor_time
	    - event: boolean indicating if failure event was observed (TRUE)
	          or censored (FALSE)
    """
    sample_data = pd.DataFrame({
            'true_t': np.random.exponential((1/rate), size=N) 
            })
    ## censor observations at censor_time
    sample_data['t'] = np.minimum(sample_data['true_t'], censor_time)
    sample_data['event'] = sample_data['t'] >= sample_data['true_t']
    sample_data['sex'] = ['female' if np.random.uniform()>0.5 else 'male' for i in np.arange(N)]
    sample_data['age'] = np.random.poisson(55, N)
    sample_data['index'] = np.arange(N)
    return(sample_data)


def _make_sim_rate(df, form, coefs):
    rate_df = patsy.dmatrix(data=df, formula_like=form, return_type='matrix')
    return np.exp(np.dot(rate_df, coefs))


def sim_data_exp_correlated(N, censor_time, rate_form = '1 + age + sex', rate_coefs = [-3, 0.3, 0]):
    """
      simulate true lifetimes (t) according to exponential model

      Parameters
      -----------

        N: (int) number of observations 
        censor_time: (float) uniform censor time for each observation
        rate_form: names of variables to use when estimating rate. defaults to `'1 + age + sex'`
        rate_coefs: inputs to rate-calc (coefs used to estimate log-rate). defaults to `[-3, 0.3, 0]`


    Returns
    -------

    pandas DataFrame with N observations, and 3 columns:
      - true_t: "actual" simulated failure time
      - t: observed failure/censor time, given censor_time
      - event: boolean indicating if failure event was observed (TRUE)
            or censored (FALSE)
      - age: simulated age in years (poisson random variable, expectation = 55)
      - sex: simulated sex, as 'female' or 'male' (uniform 50/50 split)
      - rate: simulated rate value for each obs
    """
    sample_data = pd.DataFrame({
            'sex': ['female' if np.random.uniform()>0.5 else 'male' for i in np.arange(N)],
            'age': np.random.poisson(55, N),
            })

    sample_data['rate'] = _make_sim_rate(df=sample_data, form=rate_form, coefs=rate_coefs)
    ## censor observations at censor_time
    sample_data['true_t'] = sample_data['rate'].apply(lambda rate: np.random.exponential(1/rate))
    sample_data['t'] = np.minimum(sample_data['true_t'], censor_time)
    sample_data['event'] = sample_data['t'] >= sample_data['true_t']
    sample_data['index'] = np.arange(N)
    return(sample_data)




import os
from fnmatch import fnmatch
import ntpath
import pkg_resources
import seaborn as sb
import matplotlib.pyplot as plt
from survivalstan import extract_baseline_hazard
import pandas as pd
import numpy as np
from lifelines.utils import survival_table_from_events
import logging
import re
import matplotlib.patches as mpatches

logger = logging.getLogger(__name__)

def _summarize_survival(df, time_col, event_col, evaluate_at=None):
    ## prepare survival table
    table = survival_table_from_events(df[time_col], df[event_col])
    table.reset_index(inplace=True)
    ## normalize survival as fraction of initial_n
    table['initial_n'] = table.loc[table['event_at'] == 0.0,'at_risk'][0]
    table['survival'] = table.apply(lambda row: row['at_risk']/row['initial_n'], axis=1)
    ## handle timepoints if given
    if evaluate_at is not None:
        evaluate_times = pd.DataFrame({'event_at': evaluate_at})
        table = pd.merge(evaluate_times, table, on='event_at', how='outer')
        table = table.sort_values('event_at').fillna(method='ffill')
        table['keep'] = table['event_at'].apply(lambda x: x in evaluate_at)
    else:
        table['keep'] = True
    table = table.loc[table['keep'] == True,['event_at','survival']]
    table.rename(columns={'event_at': time_col}, inplace=True)
    return table


def extract_time_betas(models, element='beta_time', value_name='beta', **kwargs):
    data = [_extract_time_betas_single_model(model, element=element, value_name=value_name, **kwargs) for model in models]
    return pd.concat(data)

def _extract_time_betas_single_model(stanmodel, element='beta_time', coefs=None,
                                     value_name='beta', timepoint_id_col=None,
                                     timepoint_end_col=None):
    if not timepoint_id_col:
        timepoint_id_col = stanmodel['timepoint_id_col']
    if not timepoint_end_col:
        timepoint_end_col = stanmodel['timepoint_end_col']
    if not timepoint_id_col or not timepoint_end_col:
        raise ValueError('timepoint_id_col and timepoint_end_col are required, but were either not given or were not set by stan model')
    time_betas = stanmodel['fit'].extract()[element]
    
    # determine coef names
    coef_names = list(stanmodel['x_names'])
    num_coefs = time_betas.shape[1]
    if len(coef_names) != num_coefs:
        raise ValueError('Num coefs does not equal number of coef names. Please report this as a bug')
    logger.debug('num_coefs set to {}'.format(num_coefs))
    
    # determine which coefs to extract
    plot_coefs = list(np.arange(num_coefs))
    if coefs is not None:
        plot_coefs = [val for val in plot_coefs if coef_names[val] in coefs]
    logger.debug('plot_coefs set to {}'.format(','.join(str(plot_coefs))))
    
    # extract time-betas for each coef
    time_data = list()
    for i in plot_coefs:
        tb_df = pd.DataFrame(time_betas[:,i,:])
        tb_df = pd.melt(tb_df, var_name=timepoint_id_col, value_name=value_name)
        tb_df['coef'] = coef_names[i]
        time_data.append(tb_df)
    time_data = pd.concat(time_data)
    timepoint_data = stanmodel['df'].loc[:,[timepoint_id_col, timepoint_end_col]].drop_duplicates()
    time_data = pd.merge(time_data, timepoint_data, on=timepoint_id_col)
    time_data['exp({})'.format(value_name)] = np.exp(time_data[value_name])
    time_data['model_cohort'] = stanmodel['model_cohort']
    return(time_data)

def _get_timepoint_cols(models, timepoint_id_col, timepoint_end_col):
    if not timepoint_id_col:
        timepoint_id_col = np.unique([model['timepoint_id_col'] for model in models])
        if len(timepoint_id_col)>1:
            ValueError('timepoint_id_col is not uniform for all models. Please reformat data and set timepoint_id_col manually')
        elif len(timepoint_id_col)==1:
            timepoint_id_col = timepoint_id_col[0]
    if not timepoint_end_col:
        timepoint_end_col = np.unique([model['timepoint_end_col'] for model in models])
        if len(timepoint_end_col)>1:
            ValueError('timepoint_end_col is not uniform for all models. Please reformat data and set timepoint_end_col manually')
        elif len(timepoint_end_col)==1:
            timepoint_end_col = timepoint_end_col[0]
    if not timepoint_id_col or not timepoint_end_col:
        raise ValueError('timepoint_id_col and timepoint_end_col are required, but were either not given or were not set by model')
    return (timepoint_id_col, timepoint_end_col)
    
def _plot_time_betas(models=None, df=None, element='beta_time',
                     coefs=None, y='exp(beta)', ylabel=None, 
                     timepoint_id_col=None, timepoint_end_col=None,
                     x='timepoint_end_col', xlabel='time', 
                     subplot=None, ticks_at=None, num_ticks=10, step_size=None,
                     fill=True, value_name='beta', ylim=None, **kwargs):
    if df is None:
        df = extract_time_betas(models=models, element=element, coefs=coefs,
                                     value_name=value_name, timepoint_id_col=timepoint_id_col,
                                     timepoint_end_col=timepoint_end_col)
        timepoint_id_col, timepoint_end_col = _get_timepoint_cols(models=models,
                                                              timepoint_id_col=timepoint_id_col,
                                                              timepoint_end_col=timepoint_end_col)
        logger.debug('timepoint_id_col set to {}'.format(timepoint_id_col))
        logger.debug('timepoint_end_col set to {}'.format(timepoint_end_col))
    else:
        if not timepoint_id_col:
            timepoint_id_col = 'timepoint_id'
    if x == 'timepoint_end_col':
        time_col = timepoint_end_col
    elif x == 'timepoint_id_col':
        time_col = timepoint_id_col
    else:
        time_col = x
    logger.debug('time_col set to {}'.format(time_col))
    if not time_col:
        raise ValueError('time_col is not defined - specify name of column using `x`')
    
    if not ylabel:
        if not coefs or len(coefs)>1:
            ylabel = '{}'.format(y)
        else:
            ylabel = '{} for {}'.format(y, coefs[0])
    df.sort_values(time_col, inplace=True)
    if not subplot:
        f, ax = plt.subplots(1, 1)
    else:
        f, ax = subplot
    if ticks_at is None:
        x_min = min(df[time_col].drop_duplicates())
        x_max = max(df[time_col].drop_duplicates())
        if step_size is None:
            step_size = (x_max - x_min)/num_ticks
        ticks_at = np.arange(start=x_min, stop=x_max, step=step_size)
    time_beta_plot = df.boxplot(
        column=y,
        by=time_col,
        whis=[2.5, 97.5],
        positions=df[time_col].drop_duplicates(),
        ax=ax,
        return_type='dict',
        showcaps=False,
        patch_artist=fill,
    )
    f.suptitle('')
    _ = plt.xticks(rotation="vertical")
    _ = plt.xlabel(xlabel)
    _ = plt.ylabel(ylabel)
    _ = plt.title('')
    _ = ax.xaxis.set_ticks(ticks_at)
    _ = ax.xaxis.set_ticklabels(
         [r"%d" % (int(round(x))) for x in ticks_at])
    
    if ylim:
        _ = plt.ylim(ylim)

    if dict(**kwargs):
        _ = plt.setp(time_beta_plot[y]['boxes'], **kwargs)
        _ = plt.setp(time_beta_plot[y]['medians'], **kwargs)
        _ = plt.setp(time_beta_plot[y]['whiskers'], **kwargs)

def plot_time_betas(models=None, df=None, element='beta_time',
                    y='beta', trans=None, coefs=None, x='timepoint_end_col',
                    by=['model_cohort','coef'], timepoint_id_col=None, timepoint_end_col=None, 
                    subplot=None, ticks_at=None, ylabel=None, xlabel='time',
                    num_ticks=10, step_size=None, fill=True, alpha=0.5, pal=None,
                    value_name='beta', **kwargs):
    if df is None:
        df = extract_time_betas(models=models, element=element, coefs=coefs,
                                     value_name=value_name, timepoint_id_col=timepoint_id_col,
                                     timepoint_end_col=timepoint_end_col)
        timepoint_id_col, timepoint_end_col = _get_timepoint_cols(models=models,
                                                              timepoint_id_col=timepoint_id_col,
                                                              timepoint_end_col=timepoint_end_col)
        logger.debug('timepoint_id_col set to {}'.format(timepoint_id_col))
        logger.debug('timepoint_end_col set to {}'.format(timepoint_end_col))
    if trans:
        trans_name = '{}({})'.format(trans.__name__, y)
        df[trans_name] = trans(df[y])
        y = trans_name
    if by:
        if not pal:
            num_grps = len(df.drop_duplicates(subset=by).loc[:, by].values)
            pal = _get_color_palette(num_grps)
        legend_handles = list()
        i = 0
        if not subplot:
            subplot = plt.subplots(1, 1)
        for grp, grp_df in df.groupby(by):
            _plot_time_betas(df=grp_df.copy(),
                             timepoint_id_col=timepoint_id_col, timepoint_end_col=timepoint_end_col,
                             num_ticks=num_ticks, step_size=step_size, ticks_at=ticks_at,
                             x=x, y=y, color=pal[i], subplot=subplot, alpha=alpha, fill=fill, **kwargs)
            legend_handles.append(mpatches.Patch(color=pal[i], label=grp))
            i = i+1
        plt.legend(handles=legend_handles)
        plt.show()
    else:
        _plot_time_betas(df=df, num_ticks=num_ticks, step_size=step_size, ticks_at=ticks_at,
                         timepoint_id_col=timepoint_id_col, timepoint_end_col=timepoint_end_col,
                         x=x, y=y, subplot=subplot, alpha=alpha, fill=fill, **kwargs)

def _get_sample_ids_single_model(model, sample_col=None, sample_id_col=None):
    if not sample_col:
        sample_col = model['sample_col']
        if not sample_col:
            raise ValueError('sample_col was not given and is also not provided to the model. This is a required input')
    if not sample_id_col:
        sample_id_col = model['sample_id_col']
        if not sample_id_col:
            raise ValueError('sample_id_col was not given and is also not provided to the model. This is a required input')
    patient_sample_ids = model['df'].loc[:,[sample_col, sample_id_col]].drop_duplicates().sort_values(sample_id_col)
    patient_sample_ids['model_cohort'] = model['model_cohort']
    return patient_sample_ids


def get_sample_ids(models, sample_col='patient_id'):
    data = [_get_sample_ids_single_model(model=model, sample_col=sample_col) for model in models]
    return pd.concat(data)


def _prep_pp_data_single_model(model, time_element='y_hat_time', event_element='y_hat_event',
                               sample_col=None, time_col='event_time', event_col='event_status',
                               join_with='df_all', sample_id_col=None):
    patient_sample_ids = _get_sample_ids_single_model(model=model, sample_col=sample_col, sample_id_col=sample_id_col)
    if not sample_col:
        sample_col = model['sample_col']
    pp_event_time = extract_params_long(models=[model],
                                       element=time_element,
                                       varnames=patient_sample_ids[sample_col].values,
                                       )
    pp_event_time.rename(columns={'value': time_col, 'variable': sample_col}, inplace=True)
    pp_event_status = extract_params_long(models=[model],
                                       element=event_element,
                                       varnames=patient_sample_ids[sample_col].values,
                                       )
    pp_event_status.rename(columns={'value': event_col, 'variable': sample_col}, inplace=True)
    pp_data = pd.merge(pp_event_time, pp_event_status, on=['iter', sample_col, 'model_cohort'])
    if join_with:
            pp_data = pd.merge(pp_data, model[join_with], on=sample_col, suffixes=['','_original'])
    return pp_data


def prep_pp_data(models, time_element='y_hat_time', event_element='y_hat_event', event_col='event_status', time_col='event_time', **kwargs):
    data = [_prep_pp_data_single_model(model=model, event_element=event_element, time_element=time_element, event_col=event_col, time_col=time_col, **kwargs)
            for model in models]
    data = pd.concat(data)
    data.sort_values([time_col, 'iter'], inplace=True)
    return data


def prep_pp_survival_data(models, time_element='y_hat_time', event_element='y_hat_event',
                         time_col='event_time', event_col='event_status',
                         by=None,
                         **kwargs):
    pp_data = prep_pp_data(models, time_element=time_element, event_element=event_element, time_col=time_col, event_col=event_col, **kwargs)
    groups = ['iter', 'model_cohort']
    if by:
        groups.append(by)
    pp_surv = pp_data.groupby(groups).apply(
         lambda df: _summarize_survival(df, time_col=time_col, event_col=event_col))
    pp_surv.reset_index(inplace=True)
    return pp_surv

def _plot_pp_survival_data(pp_surv, time_col='event_time', survival_col='survival',
                           num_ticks=10, step_size=None, ticks_at=None, subplot=None,
                           ylabel='Survival %', xlabel='Days', label='posterior predictions',
                           fill=True, **kwargs):
    pp_surv.sort_values(time_col, inplace=True)
    if not subplot:
        f, ax = plt.subplots(1, 1)
    else:
        f, ax = subplot
    if ticks_at is None:
        x_min = min(pp_surv[time_col].drop_duplicates())
        x_max = max(pp_surv[time_col].drop_duplicates())
        if step_size is None:
            step_size = (x_max - x_min)/num_ticks
        ticks_at = np.arange(start=x_min, stop=x_max, step=step_size)
    survival_plot = pp_surv.boxplot(
        column=survival_col,
        by=time_col,
        whis=[2.5, 97.5],
        positions=pp_surv[time_col].drop_duplicates(),
        ax=ax,
        return_type='dict',
        showcaps=False,
        patch_artist=fill,
    )
    f.suptitle('')
    _ = plt.ylim([0, 1])
    _ = plt.xticks(rotation="vertical")
    _ = plt.xlabel(xlabel)
    _ = plt.ylabel(ylabel)
    _ = plt.title('')
    _ = ax.xaxis.set_ticks(ticks_at)
    _ = ax.xaxis.set_ticklabels(
         [r"%d" % (int(round(x))) for x in ticks_at])

    if dict(**kwargs):
        _ = plt.setp(survival_plot[survival_col]['boxes'], **kwargs)
        _ = plt.setp(survival_plot[survival_col]['medians'], **kwargs)
        _ = plt.setp(survival_plot[survival_col]['whiskers'], **kwargs)


def _get_color_palette(n):
    """ Pick a color palette given number of values
        Returns an array containing color inputs for each value
    """
    if n <= 4:
        color_list = plt.cm.Set1(np.linspace(0, 1, n))
    else:
        color_list = plt.cm.viridis(np.linspace(0, 1, n))
    return color_list

def plot_pp_survival(models, time_element='y_hat_time', event_element='y_hat_event',
                     num_ticks=10, step_size=None, ticks_at=None, time_col='event_time',
                     event_col='event_status', fill=True, by=None, alpha=0.5, pal=None,
                     subplot=None, **kwargs):
    pp_surv = prep_pp_survival_data(models, time_element=time_element,
                                    event_element=event_element, time_col=time_col,
                                    event_col=event_col, by=by)
    if by:
        if not pal:
            num_grps = len(pp_surv.drop_duplicates(subset=by)[by].values)
            pal = _get_color_palette(num_grps)
        legend_handles = list()
        i = 0
        if not subplot:
            subplot = plt.subplots(1, 1)
        for grp, df in pp_surv.groupby(by):
            _plot_pp_survival_data(df.copy(), num_ticks=num_ticks, step_size=step_size, ticks_at=ticks_at,
                                  time_col=time_col, color=pal[i], subplot=subplot, alpha=alpha, fill=fill, **kwargs)
            legend_handles.append(mpatches.Patch(color=pal[i], label=grp))
            i = i+1
        plt.legend(handles=legend_handles)
        plt.show()
    else:
        _plot_pp_survival_data(pp_surv, num_ticks=num_ticks, step_size=step_size,
                               ticks_at=ticks_at, time_col=time_col, alpha=alpha, fill=fill, **kwargs)


def plot_observed_survival(df, event_col, time_col, label='observed', *args, **kwargs):
    actual_surv = _summarize_survival(df=df, time_col=time_col, event_col=event_col)
    plt.plot(actual_surv[time_col], actual_surv['survival'], label=label, *args, **kwargs)

def _list_files_in_path(path, pattern="*.stan"):
    """
    indexes a directory of stan files
    returns as dictionary containing contents of files
    """

    results = []
    for dirname, subdirs, files in os.walk(path):
        for name in files:
            if fnmatch(name, pattern):
                results.append(os.path.join(dirname, name))
    return(results)


def _read_file(filepath, resource=None):
    """
    reads file contents from disk

    Parameters
    ----------
    filepath (string):
        path to file (can be relative or absolute)
    resource (string, optional):
        if given, path is relative to package install root
        used to load stan files provided by packages
        (e.g. those within a package library)

    Returns
    -------
    The specifics of the return type depend on the value of `resource`.
        - if resource is None, returns contents of file as a character string
        - otherwise, returns a "resource_string" which
            acts as a character string but technically isn't one.
    """
    print(filepath)
    if not(resource):
        with open(filepath, 'r') as myfile:
            data = myfile.read()
    else:
        data = pkg_resources.resource_string(
            resource, filepath)
    return data


def read_files(path, pattern='*.stan', encoding="utf-8", resource=None):
    """
    Reads file contents from a directory path into memory. Returns a 
    dictionary of file names: file contents.

    Is intended to be used to load a directory of stan files into an object.

    Parameters
    ----------
    path (string):
        directory path (can be relative or absolute)
    pattern (string, optional):
        regex pattern applied to files on import
        defaults to "*.stan"
    encoding (string, optional):
        encoding to use when importing files
        defaults to "UTF-8"
    resource (string, optional):
        if given, path is relative to package install root
        used to load stan files provided by packages
        (e.g. those within a package library)

    Returns
    -------
    The specifics of the return type depend on the value of `resource`.
        - if resource is None, returns contents of file as a character string
        - otherwise, returns a "resource_string" which
            acts as a character string but technically isn't one.
    """
    files = _list_files_in_path(path=path, pattern=pattern)
    results = {}
    for file in files:
        file_data = {}
        file_data['path'] = file
        file_data['basename'] = ntpath.basename(file)
        file_data['code'] = _read_file(
            file,
            resource=resource).decode(encoding)
        results[file_data['basename']] = file_data['code']
    return(results)

def _prep_data_for_coefs(models, element):
    """ 
    Helper function to concatenate/extract data 
    from a list of model objects.

    See `plot_coefs` for description of data inputs
    """

    # concatenate data from models given
    df_list = list()
    [df_list.append(model[element]) for model in models]
    df = pd.concat(df_list, ignore_index=True)
    return 'value', 'variable', df


def _prep_data_for_baseline_hazard(models, element='baseline'):
    """ 
    Helper function to concatenate/extract baseline hazard data 
    from a list of model objects.
    
    Note `element` input parameter is ignored here.

    See `plot_coefs` for description of data inputs
    """
    # prepare df containing posterior estimates of baseline hazards
    df_list = list()
    [df_list.append(extract_baseline_hazard(model, element=element)) for model in models]
    df = pd.concat(df_list)

    # add helper variables to df
    df['timepoint_id'] = df['timepoint_id'].astype('category')
    df['log_hazard'] = np.log1p(df['baseline_hazard'])
    df['end_time_id'] = df['end_time'].astype('category')
    return 'log_hazard', 'end_time_id', df


def plot_coefs(models, element='coefs', force_direction=None, trans=None, **kwargs):
    """
    Plot coefficients for models listed

    Parameters
    ----------

    models (list):
        List of model objects
    element (string, optional):
        Which element to plot. defaults to 'coefs'.
        Other options (depending on model type) include: 
        - 'grp_coefs'
        - 'baseline'
        - 'beta_time'
    force_direction (string, optional):
        Takes values 'h' or 'v'
            - if 'h': forces horizontal orientation, (`variable` names along the x axis)
            - if 'v': forces vertical orientation (`variable` names along the y axis)
        if None (default), coef plots default to 'v' for all plots except baseline hazard.
    trans (function, optional):
        If present, transforms value of `value` column
            - example: np.exp to plot exp(beta)
        if None (default), plots raw value

    """

    # TODO: check if models object is a list or a single model
    
    if element=='beta_time':
        return plot_time_betas(models=models, element=element, trans=trans, **kwargs)
    
    # prep data from models given
    if element=='baseline' or element=='baseline_raw':
        value, variable, df = _prep_data_for_baseline_hazard(models, element=element)
    else:
        value, variable, df = _prep_data_for_coefs(models=models, element=element)
    
    if trans:
        df[value] = trans(df[value])

    # select hue depending on number of elements
    if len(models)==1:
        hue = None
    else:
        hue = 'model_cohort'

    if element=='baseline' or element=='baseline_raw':
        direction = 'h'
    else:
        direction = 'v'

    if force_direction:
        direction = force_direction

    if direction=='h':
        xval = variable
        yval = value
    else:
        xval = value
        yval = variable

    ## plot coefficients
    sb.boxplot(x = xval, y = yval, data = df, hue = hue)
    if hue=='model_cohort':
        plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)


def extract_params_long(models, element, rename_vars=None, varnames=None):
    """
    Helper function to extract & reformat params

    Parameters
    ----------

    models (list):
        List of model objects
    element (string, optional):
        Which element to plot. defaults to 'coefs'.
        Other options (depending on model type) include: 
        - 'grp_coefs'
        - 'baseline_hazard'
    rename_vars (dict, optional):
        - dictionary mapping from integer positions (0, 1, 2) to variable names
    varnames (list of strings, optional):
        - list of variable names to apply to columns from the extracted object

    Returns
    -------
    Pandas dataframe containing posterior draws per iteration

    """
    df_list = list()
    for model in models:
        df_list.append(_extract_params_from_single_model(
            model,
            element = element,
            rename_vars=rename_vars,
            varnames=varnames
            ))
    df_list = pd.concat(df_list)
    return(df_list)


def _extract_params_from_single_model(model, element, rename_vars=None, varnames=None):
    if varnames is None:
        df = pd.DataFrame(
            model['fit'].extract()[element]
        )
    else: 
        df = pd.DataFrame(
            model['fit'].extract()[element]
            , columns=varnames
        )
    if rename_vars is not None:
        df.rename(columns = rename_vars, inplace=True)
    df.reset_index(0, inplace = True)
    df = df.rename(columns = {'index':'iter'})
    df = pd.melt(df, id_vars = ['iter'])
    df['model_cohort'] = model['model_cohort']
    return(df)


def filter_stan_summary(stan_fit, pars=None, remove_nan=False):
    """ Filter stan fit summary, for the set of parameters in `pars`.
        See ?pystan.summary for details about summary stats given.
    
        Parameters
        -------

        stan_fit: 
            StanFit object for which posterior draws are desired to be summarized
        pars: (list, optional)
            list of strings used to filter parameters. Passed directly to `pystan.summary`.
            default: return all parameters
        remove_nan: (bool, optional)
            whether to remove (and report on) NaN values for Rhat. These are problematic for distplot.

        Returns 
        -------
        pandas dataframe containing summary stats for posterior draws of selected parameters


    """
    if isinstance(stan_fit, list):
        if len(stan_fit)>1:
            logger.warning('More than one model passed to `filter_stan_summary`. Using only the first.')
        stan_fit = stan_fit[0]['fit']
        ## else: assume stan_fit was passed correctly
    if pars:
        fitsum = stan_fit.summary(pars=pars)
    else:
        fitsum = stan_fit.summary()
    df = pd.DataFrame(fitsum['summary'], columns=fitsum['summary_colnames'], index=fitsum['summary_rownames'])
    if remove_nan:
        ## most of NaN values are Rhat
        ## remove & report on their frequency if remove_nan == True
        df_nan_rows = pd.isnull(df).any(1)
        if any(df_nan_rows):
            logger.info('Warning - {} rows removed due to NaN values for Rhat. This may indicate a problem in your model estimation.'.format(df_nan_rows[df_nan_rows].count()))
            df = df[~df_nan_rows]
    return df.loc[:,['mean','se_mean','sd','2.5%','50%','97.5%','Rhat']]


def print_stan_summary(stan_fit, pars=None):
    """ Convenience function to print stan fit summary, for the set of parameters in `pars`.

        Parameters
        -------

        stan_fit: 
            StanFit object for which posterior draws are desired to be summarized
        pars: (optional)
            list of strings used to filter parameters. Passed directly to `pystan.summary`.
            default: return all parameters
    """
    print(filter_stan_summary(stan_fit=stan_fit, pars=pars).to_string())


def plot_stan_summary(stan_fit, pars=None, metric='Rhat'):
    """ Plot distribution of values in stan fit summary, for the set of parameters in `pars`.

        Primary use case is to summarize Rhat estimates for set of parameters, as a quick check of convergence.

        Parameters
        -------

        stan_fit: 
            StanFit object for which posterior draws are desired to be summarized
        pars: (list of str, optional)
            list of strings used to filter parameters. Passed directly to `pystan.summary`.
            default: return all parameters
        metric: (str, optional)
            the name of the metric to plot, as one of: ['mean','se_mean','sd','2.5%','50%','97.5%','Rhat']
            default: `Rhat`
    """
    df = filter_stan_summary(stan_fit=stan_fit, pars=pars, remove_nan=True)
    if not metric in df.columns:
        raise ValueError('Invalid metric ({}). Should be one of {}'.format(metric, '.'.join(df.columns)))
    sb.distplot(df[metric])



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['utils', 'models', 'sim']

import sys as _sys
if (_sys.version_info > (3, 0)):
    # Python 3 code in this block
    from .survivalstan import *
    from . import utils, models, sim
    __all__ = ['utils', 'models', 'sim']
else:
    import utils
    import models
    import sim
    from survivalstan import *

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import statsmodels
import survivalstan
import random
random.seed(9001)

def load_test_dataset(n=50):
    ''' Load test dataset from R survival package
    '''
    dataset = statsmodels.datasets.get_rdataset(package='survival', dataname='flchain' )
    d = dataset.data.query('futime > 7').sample(n=n)
    d.reset_index(level=0, inplace=True)
    d.rename(columns={'futime': 't', 'death': 'event'}, inplace=True)
    return(d)


def sim_test_dataset(n=50):
    dataset = survivalstan.sim.sim_data_exp_correlated(N=n, censor_time=10)
    return(dataset)


def load_test_dataset_long(n=20):
    ''' Load test dataset from R survival package
    '''
    d = load_test_dataset(n=n)
    dlong = survivalstan.prep_data_long_surv(d, time_col='t', event_col='event')
    return dlong

def sim_test_dataset_long(n=20):
    d = sim_test_dataset(n=n)
    dlong = survivalstan.prep_data_long_surv(d, time_col='t', event_col='event')
    return dlong




import matplotlib as mpl
mpl.use('Agg')
import survivalstan
from stancache import stancache
import numpy as np
from nose.tools import ok_
from functools import partial
num_iter = 500
from .test_datasets import load_test_dataset_long, sim_test_dataset_long, load_test_dataset

model_code = survivalstan.models.pem_survival_model
make_inits = None

def test_null_pem_model(**kwargs):
    ''' Test weibull survival model on simulated dataset
    '''
    d = load_test_dataset(n=20)
    dlong = survivalstan.prep_data_long_surv(df=d, time_col='t', event_col='event')
    testfit = survivalstan.fit_stan_survival_model(
        model_cohort = 'test model',
        model_code = model_code,
        df = dlong,
        sample_col = 'index',
        timepoint_end_col = 'end_time',
        event_col = 'end_failure',
        formula = '~ 1',
        iter = num_iter,
        chains = 2,
        seed = 9001,
        make_inits = make_inits,
        FIT_FUN = stancache.cached_stan_fit,
        **kwargs
        )
    ok_('fit' in testfit)
    ok_('coefs' in testfit)
    ok_('loo' in testfit)
    survivalstan.utils.plot_coefs([testfit])
    survivalstan.utils.plot_coefs([testfit], trans=np.exp)
    survivalstan.utils.plot_coefs([testfit], element='baseline')

    survivalstan.utils.plot_pp_survival([testfit])
    survivalstan.utils.plot_observed_survival(df=d, time_col='t', event_col='event')
    fitsum = survivalstan.utils.filter_stan_summary([testfit], pars='baseline')
    fitsum = survivalstan.utils.filter_stan_summary(testfit['fit'], remove_nan=True)
    survivalstan.utils.print_stan_summary([testfit], pars='lp__')
    survivalstan.utils.plot_stan_summary([testfit], pars='log_baseline_raw')
    return(testfit)


def test_pem_model(**kwargs):
    ''' Test weibull survival model on simulated dataset
    '''
    d = load_test_dataset(n=20)
    dlong = load_test_dataset_long()
    testfit = survivalstan.fit_stan_survival_model(
        model_cohort = 'test model',
        model_code = model_code,
        df = dlong,
        sample_col = 'index',
        timepoint_end_col = 'end_time',
        event_col = 'end_failure',
        formula = '~ age + sex',
        iter = num_iter,
        chains = 2,
        seed = 9001,
        make_inits = make_inits,
        FIT_FUN = stancache.cached_stan_fit,
        **kwargs
        )
    ok_('fit' in testfit)
    ok_('coefs' in testfit)
    ok_('loo' in testfit)
    survivalstan.utils.plot_coefs([testfit])
    survivalstan.utils.plot_coefs([testfit], trans=np.exp)
    survivalstan.utils.plot_coefs([testfit], element='baseline')
    survivalstan.utils.plot_pp_survival([testfit])
    survivalstan.utils.plot_observed_survival(df=d, time_col='t', event_col='event')
    return(testfit)




from __future__ import print_function
from setuptools import setup, find_packages
from codecs import open
from os import path
import versioneer

__version__ = '0.0.1'

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# get the dependencies and installs
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')

install_requires = [x.strip() for x in all_reqs if 'git+' not in x]
dependency_links = [x.strip().replace('git+','') for x in all_reqs if 'git+' in x]

setup(
    name='survivalstan',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Library of Stan Models for Survival Analysis',
    long_description=long_description,
    url='https://github.com/hammerlab/survivalstan',
    download_url='https://github.com/hammerlab/survivalstan/tarball/' + versioneer.get_version(),
    license="http://www.apache.org/licenses/LICENSE-2.0.html",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    keywords='',
    packages=find_packages(exclude=['docs', 'test*']),
    include_package_data=True,
    author='Jacki Novik',
    install_requires=install_requires,
    dependency_links=dependency_links,
    author_email='jackinovik@gmail.com'
)
